# -*- coding:utf-8 -*-
"""这里定义一些框架基类和元类"""
import types
import pickle

from .utils import data_cmd_wrapper, common_cmd_wrapper


class Meta(type):
    """元类基类，给方法增加装饰器"""
    wrapper = None

    def __new__(typ, name, bases, properties):
        # 一些公共继承方法继承在DataStore里面，不需要被装饰，通过其基类来判断要构造的是否为DataStore类
        if bases[0] != object:
            for k, v in properties.items():
                if isinstance(v, types.FunctionType):
                    # 由于这个方法会被继承，通过提供不同的wrapper函数来做不同的包装， RedisMeta没有提供，所以不包装
                    properties[k] = typ.wrapper(v) if typ.wrapper else v

        return super(Meta, typ).__new__(typ, name, bases, properties)


class StoreMeta(Meta):
    """数据类专用元类"""
    wrapper = staticmethod(data_cmd_wrapper)


class CommonCmdMeta(StoreMeta):
    """通用函数类专用元类"""
    wrapper = staticmethod(common_cmd_wrapper)


class RedisMeta(CommonCmdMeta):
    """Redis类专用元类"""
    wrapper = None

    def __new__(typ, *args, **kwargs):
        # 这个地方非常诡异， 如果通过type(*args)来组建类，类会使用CommonCmdMeta来创建
        # 通过type.__new__则会使用type来创建类
        # 这里我们不需要对CustomRedis类的函数进行包装操作，所以选择这种创建方式
        return type.__new__(typ, *args)


class DataStore(object, metaclass=StoreMeta):
    """数据基类"""
    data_type = None

    def __init__(self, logger, data=None):
        self.logger = logger
        self.data = data or self.data_type()

    @classmethod
    def from_redis(cls, redis):
        return cls(redis.logger)

    @staticmethod
    def format_response(code, info, data):
        if data is None:
            data = b""
        return b"%s#-*-#%s#-*-#%s\r\n\r\n" % (code, info, data)

    def persist(self, stream):
        stream.write(pickle.dumps(self.data))
        stream.write(b"fdfsafafdsfsfdsfafdff")

    def _parses(self, v):
        return pickle.loads(v)

    @classmethod
    def loads(cls, val):
        if isinstance(val, cls.data_type):
            return cls




# -*- coding:utf-8 -*-
"""
redis server
主线程使用select进程socket监听和任务处理
守护线程用来持久化数据和删除过期key
"""
import os
import sys
import time
import errno
import socket
import select
import pickle
import logging
import traceback

from logging import handlers
from functools import reduce
from argparse import ArgumentParser
from threading import Thread, RLock

from multi_thread_closing import MultiThreadClosing

from .bases import RedisMeta
from .common_command import CommonCmd
from .errors import MethodNotExist, ClientClosed
from .utils import stream_wrapper, LoggerDiscriptor
from .data_types import ZsetStore, StrStore, HashStore, SetStore, ListStore


class CustomRedis(CommonCmd, MultiThreadClosing, metaclass=RedisMeta):

    name = "redis_server"
    default = {"str": StrStore, "hash": HashStore, "set": SetStore, "zset": ZsetStore, "list": ListStore}
    logger = LoggerDiscriptor()

    def __init__(self, host, port, **kwargs):
        MultiThreadClosing.__init__(self)
        self.meta = kwargs
        self.host = host
        self.port = port
        # 所有数据类型
        self.data_type = {}
        # 所有数据实例
        self.datas = {}
        # 所有数据的过期时间
        self.expire_keys = {}
        self.lock = RLock()
        self.open()
        self.data_type.update(self.default)
        self.r_lst = {}
        self.w_lst = {}

    def set_logger(self, logger=None):
        self.logger = logging.getLogger(self.name)
        self.logger.setLevel(getattr(logging, self.meta.get("log_level", "DEBUG")))
        if self.meta.get("log_file"):
            handler = handlers.RotatingFileHandler(
                os.path.join(self.meta.get("log_dir", "."), "%s.log" % self.name),
                maxBytes=10240000, backupCount=5)
        else:
            handler = logging.StreamHandler(sys.stdout)
        formater = logging.Formatter(self.meta.get("log_format", "%(message)s"))
        handler.setFormatter(formater)
        self.logger.addHandler(handler)

    def install(self, **kwds):
        self.data_type.update(kwds)

    def setup(self):
        if os.path.exists("redis_data.db"):
            lines = open("redis_data.db", "rb").read().split(b"fdfsafafdsfsfdsfafdff")
            if lines:
                self.logger.info("load datas...")
                for line in lines:
                    self.load(line)

    def load(self, line):
        # 加载数据类型
        try:
            if line:
                key, expire_time, val = line.split(b"1qazxsw23edc")
                val = pickle.loads(val)
                if len(self.data_type) < 2:
                    cls = list(self.data_type.values())[0].loads(val)
                else:
                    cls = reduce(lambda x, y:
                                 (x.loads(val) if hasattr(x, "loads") else None) or
                                 y.loads(val),
                                 self.data_type.values())
                if cls:
                    self.datas[key] = cls(self.logger, val)
                    if expire_time != b"-1":
                        self.expire_keys[key] = int(expire_time)
        except Exception:
            self.logger.error(traceback.format_exc())

    def __getattr__(self, item):
        # 遍历所有数据类型，找到数据类型对应方法并返回
        for k, v in self.data_type.items():
            attr = getattr(v, item, None)
            if attr:
                return getattr(v.from_redis(self), item)
        raise MethodNotExist

    def start(self):
        self.setup()
        daemon_thread = Thread(target=self.poll)
        daemon_thread.setDaemon(True)
        daemon_thread.start()
        self.listen_request(self.host, self.port)

    def poll(self):
        """删除过期的Key, 删除空集合及持久化数据的守护进程"""
        t = time.time()
        while True:
            # 每30秒持久化一次数据
            if time.time()-t > 30:
                t = time.time()
                self.persist()
            for key in list(self.expire_keys.keys()):
                if self.expire_keys[key] < time.time():
                    del self.datas[key], self.expire_keys[key]
            for key in [i for i in self.datas.keys()]:
                if not self.datas[key].data:
                    del self.datas[key]
            time.sleep(1)

    def listen_request(self, host, port):
        """监听函数"""
        self.logger.info("listen  to %s:%s"%(host, port))
        server = socket.socket()
        server.bind((host, port))
        server.listen(10)
        self.r_lst[server] = None
        # 若执行过程中出现异常r_lst中的server也被清掉，程序退出
        try:
            while self.alive and self.r_lst:
                readable, writable, _ = select.select(self.r_lst.keys(), self.w_lst.keys(), [], 0.1)
                for r in readable:
                    self.recv(r, self.w_lst, self.r_lst, server)
                for w in writable:
                    self.send(w, self.w_lst, self.r_lst, None)
        except select.error as e:
            if e.args[0] != 4:
                raise
        self.persist()
        self.close()
        self.logger.info("exit...")

    def close(self):
            # 只关r_list的即可
            for i in self.r_lst.keys():
                try:
                    i.close()
                except Exception:
                    pass

    def stop(self, *args):
        self.alive = False

    def persist(self, stream=None):
        self.lock.acquire()
        with open("redis_data.db", "wb") as stream:
            self.logger.info("persist datas...")
            for key, val in self.datas.items():
                stream.write(key)
                stream.write(b'1qazxsw23edc%d' % self.expire_keys.get(key, -1))
                stream.write(b'1qazxsw23edc')
                val.persist(stream)
        self.lock.release()

    @stream_wrapper
    def send(self, w, w_lst, r_lst, server):
        # item会被保存在w_lst相应的val中
        item = w_lst[w]
        self.logger.debug("start to send item %s to %s:%s" % (item, r_lst[w][0], r_lst[w][1]))
        w.send(item)

    @stream_wrapper
    def recv(self, r, w_lst, r_lst, server):
        if r is server:
            client, adr = r.accept()
            self.logger.debug("get connection from %s:%s" % (adr[0], adr[1]))
            # 将新收到的socket设置为非阻塞， 并将其保存在r_lst中
            client.setblocking(0)
            r_lst[client] = adr
        else:
            self.logger.debug("start to recv data from %s:%s" % (r_lst[r][0], r_lst[r][1]))
            received = self._recv(r)
            if received:
                cmd, data, keep = received.split(b"#-*-#")
                cmd = cmd.decode("utf-8")
                key, val = data.split(b"<->")
                # 根据指令生成item返回结果
                try:
                    method = getattr(self.datas.get(key, None), cmd, None) or getattr(self, cmd)
                    #method = getattr(self.datas.get(key, None), cmd, None)
                    # 数据类型的方法
                    if method and key in self.datas and method.__self__ != self and \
                                    self.datas[key].__class__ != method.__self__.__class__:
                        # 类型不符合
                        item = b"503#-*-#Type Not Format#-*-#\r\n\r\n"
                    else:
                        item = method(key, val, self)
                    # else:
                    #     # 通用方法
                    #     item = getattr(self, cmd)(key, val, self)
                except MethodNotExist:
                    item = b"404#-*-#Method Not Found#-*-#\r\n\r\n"
                # 将socket保存在w_lst中，并将keep-alive 标志保存在其val中
                w_lst[r] = item
                r_lst[r] = r_lst[r][:2] + (keep,)
            else:
                raise ClientClosed("closed")

    def _recv(self, r):
        msg = b""
        try:
            buf = r.recv(1024)
            while buf:
                msg += buf
                buf = r.recv(1024)
        except Exception as e:
            # 非阻塞异常直接返回
            if e.args[0] != errno.EAGAIN:
                pass
        self.logger.info("received massage is %s" % (msg or None))
        return msg

    @classmethod
    def parse_args(cls):
        parser = ArgumentParser()
        parser.add_argument("--host", help="host", default="127.0.0.1")
        parser.add_argument("-p", "--port", type=int, help="port", default=6379)
        parser.add_argument("-lf", "--log-file", action="store_true", help="log to file, else log to stdout. " )
        parser.add_argument("-ld", "--log-dir", default=".")
        parser.add_argument("-ll", "--log-level", default="DEBUG", choices=["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL"])
        parser.add_argument("--log-format", default="%(asctime)s [%(name)s] %(levelname)s: %(message)s")
        return cls(**vars(parser.parse_args()))


def start_server():
    CustomRedis.parse_args().start()


if __name__ == "__main__":
    start_server()
    #CustomRedis("127.0.01", 6379).start()



# -*- coding:utf-8 -*-
import pickle
import random

from .errors import Empty
from .zset import SortedSet
from .bases import DataStore


class ZsetStore(DataStore):

    data_type = SortedSet

    def zadd(self, k, v, instance):
        k, v = [i for i in self._parses(v).items()][0]
        self.data.zadd(v, int(k))

    def zpop(self, k, v, instance):
        return pickle.dumps(self.data.zpop(v))

    def zcard(self, k, v, instance):
        return str(self.data.zcard).encode("utf-8")


class ListStore(DataStore):

    data_type = list

    def lpop(self, k, v, instance):
        if self.data:
            return self.data.pop(0)
        else:
            raise Empty

    def rpush(self, k, v, instance):
        self.data.append(v)

    def llen(self, k, v, instance):
        return str(len(self.data)).encode("utf-8")


class StrStore(DataStore):

    data_type = str

    def add(self, k, v, instance):
        self.data += v

    def slice(self, k, v, instance):
        return eval(b"self.data[%s]" % v)

    def set(self, k, v, instance):
        self.data = v

    def get(self, k, v, instance):
        return self.data


class SetStore(DataStore):

    data_type = set

    def sadd(self, k, v, instance):
        self.data.add(v)

    def scard(self, k, v, instance):
        return str(len(self.data)).encode("utf-8")

    def smembers(self, k, v, instance):
        return pickle.dumps(list(self.data))

    def srem(self, k, v, instance):
        values = self._parses(v)
        for value in values:
            self.data.remove(bytes(str(value), encoding="utf-8"))

    def sismember(self, k, v, instance):
        return str(v in self.data).encode("utf-8")

    def srchoice(self, k, v, instance):
        return random.choice(list(self.data))


class HashStore(DataStore):

    data_type = dict

    def hset(self, k, v, instance):
        self.data.update(self._parses(v))

    def hget(self, k, v, instance):
        if isinstance(v, bytes):
            v = v.decode("utf-8")
        data = self.data[v]
        if not isinstance(data, bytes):
            data = str(data).encode("utf-8")
        return data

    def hmset(self, k, v, instance):
        k_vs = self._parses(v)
        self.data.update(dict(k_vs))

    def hmget(self, k, v, instance):
        ks = self._parses(v)
        return pickle.dumps(dict(filter(lambda x: x[0] in ks, self.data.items())))

    def hgetall(self, k, v, instance):
        return pickle.dumps(self.data)

    def hincrby(self, k, v, instance):
        k_vs = self._parses(v)
        k = list(k_vs.keys())[0]
        v = k_vs[k]
        self.data[k] = int(self.data.get(k, 0)) + int(v)






# -*- coding:utf-8 -*-
"""这里定义的是一些通用的方法"""
import time
import pickle
import fnmatch

from .bases import DataStore, CommonCmdMeta


class CommonCmd(DataStore, metaclass=CommonCmdMeta):
    """通用函数类"""
    expire_keys = {}

    def keys(self, k, v, instance):
        return b"%s#-*-#%s#-*-#%s\r\n\r\n" % (b"200", b"success",
                                             pickle.dumps([x for x in self.datas.keys() if fnmatch.fnmatch(x, k)]))

    def expire(self, k, v, instance):
        if k in self.datas:
            self.expire_keys[k] = int(time.time() + int(v))
            return b"200#-*-#success#-*-#\r\n\r\n"
        raise KeyError(k)

    def type(self, k, v, instance):
        data = self.datas[k]
        return ("200#-*-#success#-*-#%s\r\n\r\n"%data.__class__.__name__[:-5].lower()).encode("utf-8")

    def ttl(self, k, v, instance):
        expire = self.expire_keys.get(k)
        if expire:
            expire = int(expire - time.time())
        else:
            expire = -1
        return ("200#-*-#success#-*-#%d\r\n\r\n" % expire).encode("utf-8")

    def delete(self, k, v, instance):
        try:
            del self.datas[k]
        except KeyError:
            pass
        return b"200#-*-#success#-*-#\r\n\r\n"

    def flushall(self, k, v, instance):
        self.datas = {}
        return b"200#-*-#success#-*-#\r\n\r\n"



# -*- coding:utf-8 -*-
from .client import start_client
from .server import start_server


VERSION = '3.1.0'

AUTHOR = "cn"

AUTHOR_EMAIL = "308299269@qq.com"

URL = "https://www.github.com/ShichaoMa/custom_redis"



# -*- coding:utf-8 -*-
from .utils import safe_loads, safe_dumps


CMD_DICT = {
    "set": {
        "args": ["name", "value"],
        "send":lambda *args: (args[0], args[1]),
    },
    "get": {
        "args": ["name",],
    },
    "hget":{
        "args":["name", "key"],
        "send":lambda *args: (args[0], args[1]),
    },
    "hset":{
        "args": ["name", "key", "value"],
        "send":lambda *args: (args[0], safe_dumps(dict([args[1:]]))),
    },
    "hmset": {
        "args": ["name", "mapping"],
        "send":lambda *args: (args[0], safe_dumps(args[1])),
    },
    "hmget": {
        "args": ["name", "key"],
        "send": lambda *args: (args[0], safe_dumps(args[1:])),
        "recv": lambda data:safe_loads(data),
    },
    "hgetall": {
        "args": ["name"],
        "recv": lambda data: safe_loads(data),
    },
    "hincrby": {
        "args": ["name", "key", "value"],
        "send": lambda *args: (args[0], safe_dumps(dict([args[1:]]))),
        "default":[1]
    },
    "zadd": {
        "args": ["name", "value", "key"],
        "send": lambda *args: (args[0], safe_dumps(dict([reversed(args[1:])]))),
    },
    "zpop": {
        "args": ["name", "withscore"],
        "send": lambda *args: (args[0], (True if args[1] in [True, "withscore"] else b"")),
        "recv": lambda data:safe_loads(data),
        "default":[False]
    },
    "zcard": {
        "args": ["name"],
        "recv": lambda x:int(x),
        "result": 0
    },
    "lpop": {
        "args": ["name"],
    },
    "rpush": {
        "args": ["name", "value"],
        "send": lambda *args: (args[0], args[1]),
    },
    "llen": {
        "args": ["name"],
        "recv": lambda x: int(x),
        "result": 0
    },
    "scard": {
        "args": ["name"],
        "recv": lambda x: int(x),
        "result": 0
    },
    "sadd": {
        "args": ["name", "value"],
        "send": lambda *args: (args[0], args[1]),
    },
    "srem": {
        "args": ["name", "values"],
        "send": lambda *args: (args[0], safe_dumps(args[1:])),
    },
    "srchoice": {
        "args": ["name"],
    },
    "smembers": {
        "args": ["name"],
        "recv": lambda x: set(safe_loads(x)),
    },
    "sismember": {
        "args": ["name", "value"],
        "send": lambda *args: (args[0], args[1]),
        "recv": lambda x: eval(x),
    },
}
